function Results = HMXL_FSE(INPUT,Results_old,EstimOpt,OptimOpt)


global B_backup;

% save tmp_HMXL
% return

tic

Results.bhat = [];
Results.R = [];
Results.R_out = {};
Results.stats = [];


%% Check data and inputs


if nargin < 3 % check no. of inputs
    error('Too few input arguments for HMXL(INPUT,EstimOpt,OptimOpt)')
end

disp(' ');
disp('__________________________________________________________________________________________________________________');
disp(' ');

warning off MATLAB:mir_warning_maybe_uninitialized_temporary

format shortG;
format compact;

if any(INPUT.W ~= 1)
    cprintf('Black','Estimating '); cprintf('*Black','weighted '); cprintf('Black','HMXL model...\n');
else
    disp('Estimating HMXL model ...')
end

if isfield(EstimOpt,'FullCov') == 0
    EstimOpt.FullCov = 0;
end
if EstimOpt.FullCov == 0
    disp('with non-correlated random parameters ...')
elseif EstimOpt.FullCov == 1
    disp('with correlated random parameters ...')
elseif EstimOpt.FullCov == 2
    disp('with correlated random parameters and latent variables ...')
end
EstimOpt.NVarA1 = size(INPUT.Xa1,2); % DCE
EstimOpt.NVarA2 = size(INPUT.Xa2,2); % FSE

if isfield(EstimOpt,'Dist1') == 0 || isempty(EstimOpt.Dist1)
    EstimOpt.Dist1 = zeros(1,EstimOpt.NVarA1);
    cprintf(rgb('DarkOrange'),'WARNING: distributions for random parameters not specified - assuming normality (Dist1) \n')
else
    if length(EstimOpt.Dist1) == 1
        EstimOpt.Dist1 = EstimOpt.Dist1.*ones(1,EstimOpt.NVarA1);
    elseif length(EstimOpt.Dist1) == EstimOpt.NVarA1
        EstimOpt.Dist1 = EstimOpt.Dist1(:)';
    else
        error('Incorrect no. of random parameters'' distributions provided (Dist1)')
    end
end
if isfield(EstimOpt,'Dist2') == 0 || isempty(EstimOpt.Dist2)
    EstimOpt.Dist2 = zeros(1,EstimOpt.NVarA2);
    cprintf(rgb('DarkOrange'),'WARNING: distributions for random parameters not specified - assuming normality (Dist2) \n')
else
    if length(EstimOpt.Dist2) == 1
        EstimOpt.Dist2 = EstimOpt.Dist2.*ones(1,EstimOpt.NVarA2);
    elseif length(EstimOpt.Dist2) == EstimOpt.NVarA2
        EstimOpt.Dist2 = EstimOpt.Dist2(:)';
    else
        error('Incorrect no. of random parameters'' distributions provided (Dist2)')
    end
end

disp(['Random parameters distributions: ',num2str(EstimOpt.Dist1),' (-1 - constant, 0 - normal, 1 - lognormal, 2 - spike, 5 - Weibull)'])

EstimOpt.NVarRnd1 = sum(EstimOpt.Dist1 ~= -1);
EstimOpt.NVarRnd2 = sum(EstimOpt.Dist2 ~= -1);
% TO DO:
    % Xm, Xstr?

if ~isfield(INPUT,'Xstr')  
    INPUT.Xstr = [];
end

EstimOpt.NVarStr = size(INPUT.Xstr,2);
if ~isfield(EstimOpt,'Type')
    EstimOpt.Type = 0; % Assuming OLS
    % Type = 0 is OLS, Type = 2 is MNL
    disp('Assuming OLS for the second model')
end

if EstimOpt.Type == 0
    EstimOpt.CutOff = 1; % sigma
    EstimOpt.NAlt2 = 1;
elseif EstimOpt.Type == 1
    U = unique(INPUT.Y2(:,1));
    EstimOpt.CutOff = length(U)-1; % no. of cutt-offs
    EstimOpt.NAlt2 = 1;
    if min(U) ~= 1
        error('Ordered response variable needs to start with 1')
    end
    if any(U(2:end)-U(1:end-1) ~= 1)
        error('For ordered responses the differences between levels should be 1')
    end
elseif EstimOpt.Type == 2 % MNL
    EstimOpt.CutOff = 0;
end

if isfield(EstimOpt,'FSEMat') == 0 || (size(EstimOpt.FSEMat,1)~= EstimOpt.NVarRnd2) || size(EstimOpt.FSEMat,2)~= EstimOpt.NVarA1
    EstimOpt.FSEMat = ones(EstimOpt.NVarRnd2, EstimOpt.NVarA1);
    disp('Assuming each latent factor explains each marginal utility')
end
EstimOpt.InterEffects = sum(sum(EstimOpt.FSEMat));

if isfield(EstimOpt,'NamesA1') == 0 || isempty(EstimOpt.NamesA1) || length(EstimOpt.NamesA1) ~= EstimOpt.NVarA1
    EstimOpt.NamesA1 = (1:EstimOpt.NVarA1)';
    EstimOpt.NamesA1 = cellstr(num2str(EstimOpt.NamesA1));
elseif size(EstimOpt.NamesA1,1) ~= EstimOpt.NVarA1
    EstimOpt.NamesA1 = EstimOpt.NamesA1';
end
if isfield(EstimOpt,'NamesA2') == 0 || isempty(EstimOpt.NamesA2) || length(EstimOpt.NamesA2) ~= EstimOpt.NVarA2
    EstimOpt.NamesA2 = (1:EstimOpt.NVarA2)';
    EstimOpt.NamesA2 = cellstr(num2str(EstimOpt.NamesA2));
elseif size(EstimOpt.NamesA2,1) ~= EstimOpt.NVarA2
    EstimOpt.NamesA2 = EstimOpt.NamesA2';
end

if isfield(EstimOpt,'NamesStr') == 0 || isempty(EstimOpt.NamesStr) || length(EstimOpt.NamesStr) ~= EstimOpt.NVarStr
    EstimOpt.NamesStr = (1:EstimOpt.NVarStr)';
    EstimOpt.NamesStr = cellstr(num2str(EstimOpt.NamesStr));
elseif size(EstimOpt.NamesStr,1) ~= EstimOpt.NVarStr
    EstimOpt.NamesStr = EstimOpt.NamesStr';
end

EstimOpt.NumGrad = 1;
disp('Setting gradient to numerical')

%% Rescructure data

INPUT2.XXa1 = reshape(INPUT.Xa1,[EstimOpt.NAlt1*EstimOpt.NCT1,EstimOpt.NP,EstimOpt.NVarA1]);
INPUT2.XXa1 = permute(INPUT2.XXa1,[1 3 2]);
INPUT2.YY1 = reshape(INPUT.Y1,[EstimOpt.NAlt1*EstimOpt.NCT1,EstimOpt.NP]);
INPUT2.Xstr = INPUT.Xstr; % Assume it is NP x NVarStr
if EstimOpt.Type < 2
     INPUT2.XXa2f = INPUT.Xa2(:,EstimOpt.Dist2 == -1);
    INPUT2.XXa2r = reshape(INPUT.Xa2(:,EstimOpt.Dist2 ~= -1),[EstimOpt.NCT2,EstimOpt.NP,EstimOpt.NVarRnd2]);
     INPUT2.XXa2r = permute(INPUT2.XXa2r,[1 3 2]);
    
    INPUT2.YY2 = reshape(INPUT.Y2,[EstimOpt.NCT2,EstimOpt.NP]);
end


% EstimOpt.indx1 = [];
% EstimOpt.indx2 = [];
% EstimOpt.indx3 = [];
% if EstimOpt.NumGrad == 0 && EstimOpt.FullCov > 0
%     for i = 1:EstimOpt.NVarA
%         EstimOpt.indx1 = [EstimOpt.indx1,i:EstimOpt.NVarA];
%         EstimOpt.indx2 = [EstimOpt.indx2,i*ones(1,EstimOpt.NVarA+1-i)];
%     end
%     if EstimOpt.FullCov == 2
%         aa = 1;
%         ab = sum(1:EstimOpt.NVarA);
%         for i = 1:EstimOpt.NVarA
%             EstimOpt.indx3 = [EstimOpt.indx3,aa:(aa+EstimOpt.NVarA-i),ab+i];
%             aa = aa+EstimOpt.NVarA-i+1;
%         end
%     end
% end



%% Starting values
if EstimOpt.FullCov == 0
    if exist('B_backup','var') && ~isempty(B_backup) && size(B_backup,1) == (EstimOpt.NVarA1 + EstimOpt.NVarRnd1 + EstimOpt.NVarA2 + EstimOpt.NVarRnd2*(1+EstimOpt.NVarStr) + EstimOpt.InterEffects + EstimOpt.CutOff)
        b0 = B_backup(:);
        disp('Using the starting values from Backup')
    elseif isfield(Results_old,'HMXL_FSE_d') && isfield(Results_old.HMXL_FSE_d,'b0') % starting values provided
        Results_old.HMXL_FSE_d.b0_old = Results_old.HMXL_FSE_d.b0(:);
        Results_old.HMXL_FSE_d = rmfield(Results_old.HMXL_FSE_d,'b0');
        if length(Results_old.HMXL_FSE_d.b0_old) ~= (EstimOpt.NVarA1 + EstimOpt.NVarRnd1 + EstimOpt.NVarA2 + EstimOpt.NVarRnd2*(1+EstimOpt.NVarStr) + EstimOpt.InterEffects + EstimOpt.CutOff)
            cprintf(rgb('DarkOrange'),'WARNING:  Incorrect no. of starting values or model specification \n')
            Results_old.HMXL_FSE_d = rmfield(Results_old.HMXL_FSE_d,'b0_old');
        else
            b0 = Results_old.HMXL_FSE_d.b0_old(:);
        end
    end
    if ~exist('b0','var')
        if isfield(Results_old,'MXL_d') && isfield(Results_old.MXL_d,'bhat')
            %% ? 
        else
            error('No starting values available - run HMNL first')
        end
    end
elseif EstimOpt.FullCov == 1
    EstimOpt.NVarRndCorr = sum(1: EstimOpt.NVarRnd1);
    if exist('B_backup','var') && ~isempty(B_backup) && size(B_backup,1) == (EstimOpt.NVarA1 + EstimOpt.NVarRndCorr + EstimOpt.NVarA2 + EstimOpt.NVarRnd2*(1+EstimOpt.NVarStr) + EstimOpt.InterEffects + EstimOpt.CutOff)
        b0 = B_backup(:);
        disp('Using the starting values from Backup')
    elseif isfield(Results_old,'HMXL_FSE') && isfield(Results_old.HMXL_FSE,'b0') % starting values provided
        Results_old.HMXL_FSE.b0_old = Results_old.HMXL_FSE.b0(:);
        Results_old.HMXL_FSE = rmfield(Results_old.HMXL_FSE,'b0');
        if length(Results_old.HMXL_FSE.b0_old) ~= (EstimOpt.NVarA1 + EstimOpt.NVarRndCorr + EstimOpt.NVarA2 + EstimOpt.NVarRnd2*(1+EstimOpt.NVarStr) + EstimOpt.InterEffects + EstimOpt.CutOff)
            cprintf(rgb('DarkOrange'),'WARNING:  Incorrect no. of starting values or model specification \n')
            Results_old.HMXL_FSE = rmfield(Results_old.HMXL_FSE,'b0_old');
        else
            b0 = Results_old.HMXL_FSE.b0_old(:);
        end
    end
    if ~exist('b0','var')
        if isfield(Results_old,'HMXL_FSE_d') && isfield(Results_old.HMXL_FSE_d,'bhat')
            disp('Using HMXL_FSE_d results as starting values')
            Results_old.HMXL_FSE_d.bhat = Results_old.HMXL_FSE_d.bhat(:);
            
            vc_tmp = abs(diag(Results_old.HMXL_FSE_d.bhat(EstimOpt.NVarA1+1:EstimOpt.NVarA1+EstimOpt.NVarRnd1)));
            b0 = [Results_old.HMXL_FSE_d.bhat(1:EstimOpt.NVarA1);vc_tmp(tril(ones(size(vc_tmp))) == 1);Results_old.HMXL_FSE_d.bhat(EstimOpt.NVarA1+EstimOpt.NVarRnd1+1:end)];
        else
            error('No starting values available - run HMXL_FSE_d first')
        end
    end
elseif EstimOpt.FullCov == 2
    EstimOpt.NVarRndCorr = sum(1: EstimOpt.NVarRnd1);
    EstimOpt.NVarRndCorr2 = sum(1: EstimOpt.NVarRnd2);
    
    if exist('B_backup','var') && ~isempty(B_backup) && size(B_backup,1) == (EstimOpt.NVarA1 + EstimOpt.NVarRndCorr + EstimOpt.NVarA2 + EstimOpt.NVarRndCorr2 + EstimOpt.NVarStr*EstimOpt.NVarRnd2 + EstimOpt.InterEffects + EstimOpt.CutOff)
        b0 = B_backup(:);
        disp('Using the starting values from Backup')
    elseif isfield(Results_old,'HMXL_FSE_c') && isfield(Results_old.HMXL_FSE_c,'b0') % starting values provided
        Results_old.HMXL_FSE_c.b0_old = Results_old.HMXL_FSE_c.b0(:);
        Results_old.HMXL_FSE_c = rmfield(Results_old.HMXL_FSE_c,'b0');
        if length(Results_old.HMXL_FSE_c.b0_old) ~= (EstimOpt.NVarA1 + EstimOpt.NVarRndCorr + EstimOpt.NVarA2 + EstimOpt.NVarRndCorr2 + EstimOpt.NVarStr*EstimOpt.NVarRnd2+ EstimOpt.InterEffects + EstimOpt.CutOff)
            cprintf(rgb('DarkOrange'),'WARNING:  Incorrect no. of starting values or model specification \n')
            Results_old.HMXL_FSE_c = rmfield(Results_old.HMXL_FSE_c,'b0_old');
        else
            b0 = Results_old.HMXL_FSE_c.b0_old(:);
        end
    end
    if ~exist('b0','var')
        if isfield(Results_old,'HMXL_FSE') && isfield(Results_old.HMXL_FSE,'bhat')
            disp('Using HMXL_FSE results as starting values')
            Results_old.HMXL_FSE.bhat = Results_old.HMXL_FSE.bhat(:);
            Indx = EstimOpt.NVarA1 + EstimOpt.NVarRndCorr + EstimOpt.NVarA2;
            vc_tmp = abs(diag(Results_old.HMXL_FSE.bhat(Indx+1:Indx+EstimOpt.NVarRnd2)));
            b0 = [Results_old.HMXL_FSE.bhat(1:Indx);vc_tmp(tril(ones(size(vc_tmp))) == 1);Results_old.HMXL_FSE.bhat(Indx+EstimOpt.NVarRnd2+1:end)];
        else
            error('No starting values available - run HMXL_FSE first')
        end
    end
end

%% Optimization Options

if  isfield(EstimOpt,'BActive')
    EstimOpt.BActive = EstimOpt.BActive(:)';
end

if EstimOpt.ConstVarActive == 1
    if ~isfield(EstimOpt,'BActive') || isempty(EstimOpt.BActive) || sum(EstimOpt.BActive == 0) == 0
        error('Are there any constraints on model parameters (EstimOpt.ConstVarActive)? Constraints not provided (EstimOpt.BActive).')
    elseif length(b0) ~= length(EstimOpt.BActive)
        error('Check no. of constraints')
    end
    disp(['Initial values: ' mat2str(b0',2)])
    disp(['Parameters with zeros are constrained to their initial values: ' mat2str(EstimOpt.BActive')])
else
    if ~isfield(EstimOpt,'BActive') || isempty(EstimOpt.BActive) || sum(EstimOpt.BActive == 0) == 0
        EstimOpt.BActive = ones(1,length(b0));
        disp(['Initial values: ' mat2str(b0',2)])
    else
        if length(b0) ~= length(EstimOpt.BActive)
            error('Check no. of constraints')
        else
            disp(['Initial values: ' mat2str(b0',2)])
            disp(['Parameters with zeros are constrained to their initial values: ' mat2str(EstimOpt.BActive')])
        end
    end
end

%% Generate pseudo-random draws

if isfield(EstimOpt,'Seed1') == 1
    rng(EstimOpt.Seed1);
end
cprintf('Simulation with ');
cprintf('*blue',[num2str(EstimOpt.NRep) ' ']);

if EstimOpt.Draws == 1
    cprintf('*blue','Pseudo-random '); cprintf('draws \n');
    err_mtx = randn(EstimOpt.NP*EstimOpt.NRep,EstimOpt.NVarRnd1+EstimOpt.NVarRnd2);
elseif EstimOpt.Draws == 2 % LHS
    cprintf('*blue','Latin Hypercube Sampling '); cprintf('draws \n');
    err_mtx=lhsnorm(zeros((EstimOpt.NVarRnd1+EstimOpt.NVarRnd2)*EstimOpt.NP,1),diag(ones((EstimOpt.NVarRnd1+EstimOpt.NVarRnd2)*EstimOpt.NP,1)),EstimOpt.NRep);
    err_mtx = reshape(err_mtx,[EstimOpt.NRep*EstimOpt.NP,EstimOpt.NLatent]);
elseif EstimOpt.Draws >= 3 % Quasi random draws
    if EstimOpt.Draws == 3
        cprintf('*blue','Halton '); cprintf(['draws (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),') \n'])
        hm1 = haltonset(EstimOpt.NVarRnd1+EstimOpt.NVarRnd2,'Skip',EstimOpt.HaltonSkip,'Leap',EstimOpt.HaltonLeap);
    elseif EstimOpt.Draws == 4 % apply reverse-radix scrambling
        cprintf('*blue','Halton '); cprintf(['draws with reverse radix scrambling (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),') \n'])
        hm1 = haltonset(EstimOpt.NVarRnd1+EstimOpt.NVarRnd2,'Skip',EstimOpt.HaltonSkip,'Leap',EstimOpt.HaltonLeap);
        hm1 = scramble(hm1,'RR2');
    elseif EstimOpt.Draws == 5
        cprintf('*blue','Sobol '); cprintf(['draws (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),') \n'])
        hm1 = sobolset(EstimOpt.NVarRnd1+EstimOpt.NVarRnd2,'Skip',EstimOpt.HaltonSkip,'Leap',EstimOpt.HaltonLeap);
    elseif EstimOpt.Draws == 6
        cprintf('*blue','Sobol '); cprintf(['draws with random linear scramble and random digital shift (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),') \n'])
        hm1 = sobolset(EstimOpt.NVarRnd1+EstimOpt.NVarRnd2,'Skip',EstimOpt.HaltonSkip,'Leap',EstimOpt.HaltonLeap);
        hm1 = scramble(hm1,'MatousekAffineOwen');
    end
    
    err_mtx = net(hm1,EstimOpt.NP*EstimOpt.NRep); % this takes every point:
    clear hm1;
    
    if EstimOpt.NP*EstimOpt.NRep < 3e+7
        err_mtx = icdf('Normal',err_mtx,0,1); %to be cut down later
    else % this is for very large number of draws * variables
        for i = 1:EstimOpt.NVarRnd1+EstimOpt.NVarRnd2
            err_mtx(:,i) = icdf('Normal',err_mtx(:,i),0,1); %to be cut down later
        end
    end        
end
err_sliced = err_mtx'; % NVarA + NLatent x NRep * NP

if isfield(EstimOpt,'Drawskeep') && ~isempty(EstimOpt.Drawskeep) && EstimOpt.Drawskeep == 1
    Results.err = err_sliced;
end

%% Display Options
if ((isfield(EstimOpt, 'ConstVarActive') == 1 && EstimOpt.ConstVarActive == 1) || sum(EstimOpt.BActive == 0) > 0) && ~isequal(OptimOpt.GradObj,'on')
    cprintf(rgb('DarkOrange'),'WARNING: Setting user-supplied gradient on - otherwise parameters'' constraints will be ignored - switch to constrained optimization instead (EstimOpt.ConstVarActive = 1) \n')
    OptimOpt.GradObj = 'on';
end


if (isfield(EstimOpt, 'ConstVarActive') == 0 || EstimOpt.ConstVarActive == 0) && isequal(OptimOpt.Algorithm,'quasi-newton') && isequal(OptimOpt.Hessian,'user-supplied')
    cprintf(rgb('DarkOrange'),'WARNING: Setting user-supplied Hessian off - quasi-newton algorithm does not use it anyway \n')
    OptimOpt.Hessian = 'off';
end
if EstimOpt.RobustStd == 1 && (EstimOpt.HessEstFix == 1 || EstimOpt.HessEstFix == 2)
    EstimOpt.RobustStd = 0;
    cprintf(rgb('DarkOrange'),'WARNING: Setting off robust standard errors, they do not matter for BHHH aproximation of hessian \n')
end
cprintf('Opmization algorithm: '); cprintf('*Black',[OptimOpt.Algorithm '\n'])

if strcmp(OptimOpt.GradObj,'on')
    if EstimOpt.NumGrad == 0
        cprintf('Gradient: '); cprintf('*Black','user-supplied, analytical \n')
    else
        cprintf('Gradient: '); cprintf('*Black',['user-supplied, numerical, ' OptimOpt.FinDiffType '\n'])
    end
else
    cprintf('Gradient: '); cprintf('*Black',['built-in, ' OptimOpt.FinDiffType '\n'])
end

if isequal(OptimOpt.Algorithm,'quasi-newton')
    cprintf('Hessian: '); cprintf('*Black','off, ')
    switch EstimOpt.HessEstFix
        case 0
            cprintf('*Black','retained from optimization \n')
        case 1
            cprintf('*Black','ex-post calculated using BHHH \n')
        case 2
            cprintf('*Black','ex-post calculated using high-precision BHHH \n')
        case 3
            cprintf('*Black','ex-post calculated numerically \n')
        case 4
            cprintf('*Black','ex-post calculated analytically \n')
    end
else
    if strcmp(OptimOpt.Hessian,'user-supplied')
        if EstimOpt.ApproxHess == 1
            cprintf('Hessian: '); cprintf('*Black','user-supplied, BHHH, ')
        else
            cprintf('Hessian: '); cprintf('*Black','user-supplied, analytical, ')
        end
    else
        cprintf('Hessian: '); cprintf('*Black',['built-in, ' OptimOpt.HessUpdate ', '])
    end
    switch EstimOpt.HessEstFix
        case 0
            cprintf('*Black','retained from optimization \n')
        case 1
            cprintf('*Black','ex-post calculated using BHHH \n')
        case 2
            cprintf('*Black','ex-post calculated using high-precision BHHH \n')
        case 3
            cprintf('*Black','ex-post calculated numerically \n')
        case 4
            cprintf('*Black','ex-post calculated analytically \n')
    end
end

%% Estimation

LLfun = @(B) LL_hmxlfse_MATlike(INPUT2, err_sliced,INPUT.W,EstimOpt,OptimOpt,B);
if EstimOpt.ConstVarActive == 0
    if EstimOpt.HessEstFix == 0
        [Results.bhat,LL,Results.exitf,Results.output,Results.g,Results.hess] = fminunc(LLfun,b0,OptimOpt);
    else
        [Results.bhat,LL,Results.exitf,Results.output,Results.g] = fminunc(LLfun,b0,OptimOpt);
    end
elseif EstimOpt.ConstVarActive == 1 % equality constraints
    EstimOpt.CONS1 = diag(1-EstimOpt.BActive);
    EstimOpt.CONS1(sum(EstimOpt.CONS1,1) == 0,:) = [];
    EstimOpt.CONS2 = zeros(size(EstimOpt.CONS1,1),1);
    if EstimOpt.HessEstFix == 0
        [Results.bhat,LL,Results.exitf,Results.output,Results.lambda,Results.g,Results.hess] = fmincon(LLfun,b0,[],[],EstimOpt.CONS1,EstimOpt.CONS2,[],[],[],OptimOpt);
    else
        [Results.bhat,LL,Results.exitf,Results.output,Results.lambda,Results.g] = fmincon(LLfun,b0,[],[],EstimOpt.CONS1,EstimOpt.CONS2,[],[],[],OptimOpt);
    end
end


%% Output

Results.LL = -LL;

Results.b0_old = b0;
if EstimOpt.HessEstFix == 1
    if isequal(OptimOpt.GradObj,'on') && EstimOpt.NumGrad == 0
        [~,Results.jacobian] = LL_hmxlfse(INPUT2, err_sliced,EstimOpt,Results.bhat);
    else
        f = LL_hmxlfse(INPUT2, err_sliced,EstimOpt,Results.bhat);
        Results.jacobian = numdiff(@(B) LL_hmxlfse(INPUT2, err_sliced,EstimOpt,B),f,Results.bhat,isequal(OptimOpt.FinDiffType,'central'),EstimOpt.BActive);
    end
    Results.jacobian = INPUT.W.*Results.jacobian;
elseif EstimOpt.HessEstFix == 2
    Results.jacobian = jacobianest(@(B) INPUT.W.*LL_hmxlfse(INPUT2, err_sliced,EstimOpt,B),Results.bhat);
elseif EstimOpt.HessEstFix == 3
    Results.hess = hessian(@(B) sum(INPUT.W.*LL_hmxlfse(INPUT2, err_sliced,EstimOpt,B),1),Results.bhat);
end
if EstimOpt.HessEstFix == 1 || EstimOpt.HessEstFix == 2
    Results.hess = Results.jacobian(:,EstimOpt.BActive == 1)'*Results.jacobian(:,EstimOpt.BActive == 1);
    EstimOpt.BLimit = (sum(Results.hess) == 0);
    EstimOpt.BLimit = direcXpnd(EstimOpt.BLimit,EstimOpt.BActive);
else
    EstimOpt.BLimit = (sum(Results.hess) == 0 & EstimOpt.BActive == 1);
    EstimOpt.BActive(EstimOpt.BLimit == 1) = 0;
    Results.hess = Results.hess(EstimOpt.BActive == 1,EstimOpt.BActive == 1);
end
Results.ihess = inv(Results.hess);
Results.ihess = direcXpnd(Results.ihess,EstimOpt.BActive);
Results.ihess = direcXpnd(Results.ihess',EstimOpt.BActive);
if EstimOpt.RobustStd == 1
    if EstimOpt.NumGrad == 0
        [~,Results.jacobian] = LL_hmxlfse(INPUT2, err_sliced,EstimOpt,Results.bhat);
        Results.jacobian = Results.jacobian.*INPUT.W;
    else
        Results.LLdetailed = LL_hmxlfse(INPUT2, err_sliced,EstimOpt,Results.bhat);
        Results.jacobian = numdiff(@(B) INPUT.W.*LL_hmxlfse(INPUT2, err_sliced,EstimOpt,B),INPUT.W.*Results.LLdetailed,Results.bhat,isequal(OptimOpt.FinDiffType,'central'),EstimOpt.BActive);
    end
    RobustHess = Results.jacobian'*Results.jacobian;
    Results.ihess = Results.ihess*RobustHess*Results.ihess;
end

Results.std = sqrt(diag(Results.ihess));
Results.std(EstimOpt.BActive == 0) = NaN;
Results.std(EstimOpt.BLimit == 1) = 0;
Results.std(imag(Results.std) ~= 0) = NaN;

Results.DetailsA1(:,1) = Results.bhat(1:EstimOpt.NVarA1);
Results.DetailsA1(:,3:4) = [Results.std(1:EstimOpt.NVarA1),pv(Results.bhat(1:EstimOpt.NVarA1),Results.std(1:EstimOpt.NVarA1))];
if EstimOpt.FullCov == 0
    Results.DetailsV1 = NaN(EstimOpt.NVarA1,4);
    Results.DetailsV1(EstimOpt.Dist1 ~= -1,1) = abs(Results.bhat(EstimOpt.NVarA1+1:EstimOpt.NVarA1+EstimOpt.NVarRnd1));
    Results.DetailsV1(EstimOpt.Dist1 ~= -1,3:4) = [Results.std(EstimOpt.NVarA1+1:EstimOpt.NVarA1+EstimOpt.NVarRnd1),pv(abs(Results.bhat(EstimOpt.NVarA1+1:EstimOpt.NVarA1+EstimOpt.NVarRnd1)),Results.std(EstimOpt.NVarA1+1:EstimOpt.NVarA1+EstimOpt.NVarRnd1))];
    l = EstimOpt.NVarA1+EstimOpt.NVarRnd1; 
   
elseif EstimOpt.FullCov == 1 || EstimOpt.FullCov == 2
%     EstimOpt.NVarA = EstimOpt.NVarA1;
    Results.DetailsV1 = sdtri_fse(Results.bhat(EstimOpt.NVarA1+1:EstimOpt.NVarA1+sum(1:EstimOpt.NVarRnd1)), ...
        Results.ihess(EstimOpt.NVarA1+1:EstimOpt.NVarA1+sum(1:EstimOpt.NVarRnd1),EstimOpt.NVarA1+1:EstimOpt.NVarA1+sum(1:EstimOpt.NVarRnd1)),EstimOpt);
    Results.DetailsV1 = [Results.DetailsV1(:,1),zeros(EstimOpt.NVarRnd1,1),Results.DetailsV1(:,2:3)];
   l = EstimOpt.NVarA1+sum(1:EstimOpt.NVarRnd1); 
end
Results.DetailsA2(:,1) = Results.bhat(l+1:l+EstimOpt.NVarA2);
Results.DetailsA2(:,3:4) = [Results.std(l+1:l+EstimOpt.NVarA2),pv(Results.bhat(l+1:l+EstimOpt.NVarA2),Results.std(l+1:l+EstimOpt.NVarA2))];
if EstimOpt.FullCov < 2
    Results.DetailsV2 = NaN(EstimOpt.NVarA2,4);
    Results.DetailsV2(EstimOpt.Dist2 ~= -1,1) = abs(Results.bhat(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+EstimOpt.NVarRnd2));
    Results.DetailsV2(EstimOpt.Dist2 ~= -1,3:4) = [Results.std(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+EstimOpt.NVarRnd2),pv(abs(Results.bhat(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+EstimOpt.NVarRnd2)),Results.std(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+EstimOpt.NVarRnd2))];
    l = l+EstimOpt.NVarA2+EstimOpt.NVarRnd2;
else
    Results.DetailsV2 = NaN(EstimOpt.NVarA2,4);
    DetailsV2 = sdtri_fse(Results.bhat(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+sum(1:EstimOpt.NVarRnd2)), ...
        Results.ihess(l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+sum(1:EstimOpt.NVarRnd2),l+EstimOpt.NVarA2+1:l+EstimOpt.NVarA2+sum(1:EstimOpt.NVarRnd2)),EstimOpt);
    Results.DetailsV2(EstimOpt.Dist2 ~= -1,1) = DetailsV2(:,1);
    Results.DetailsV2(EstimOpt.Dist2 ~= -1,3:4) = DetailsV2(:,2:3);
    l = l+EstimOpt.NVarA2+sum(1:EstimOpt.NVarRnd2);
end
% Maybe change dimensions
    % NVarA x no. of latent factors - NaNs for attributes on which
    % latent variable is not loading on 

if EstimOpt.NVarStr > 0
    Results.DetailsStr(:,1) = Results.bhat(l+1:l+EstimOpt.NVarStr*EstimOpt.NVarRnd2);
    Results.DetailsStr(:,3:4) = [Results.std(l+1:l+EstimOpt.NVarStr*EstimOpt.NVarRnd2),pv(Results.bhat(l+1:l+EstimOpt.NVarStr*EstimOpt.NVarRnd2),Results.std(l+1:l+EstimOpt.NVarStr*EstimOpt.NVarRnd2))];
    l = l+EstimOpt.NVarStr*EstimOpt.NVarRnd2;
end
Results.DetailsInter = NaN(EstimOpt.NVarA1, 4, EstimOpt.NVarRnd2);
for i = 1:EstimOpt.NVarA1
    for j = 1:EstimOpt.NVarRnd2
        if EstimOpt.FSEMat(j,i) == 1
            Results.DetailsInter(i,[1 3 4],j) =  [Results.bhat(l+1),Results.std(l+1),pv(Results.bhat(l+1),Results.std(l+1))];
            l = l+1;
        end
    end
end
Results.DetailsInter = reshape(Results.DetailsInter, [EstimOpt.NVarA1, 4*EstimOpt.NVarRnd2]);
%     Results.DetailsInter(:,1) = Results.bhat(l+1:l+EstimOpt.InterEffects);
%     Results.DetailsInter(:,3:4) = [Results.std(l+1:l+EstimOpt.InterEffects),pv(Results.bhat(l+1:l+EstimOpt.InterEffects),Results.std(l+1:l+EstimOpt.InterEffects))];
%     l = l+EstimOpt.InterEffects;
Results.DetailsCutOffs(:,1) = Results.bhat(l+1:l+EstimOpt.CutOff);
Results.DetailsCutOffs(:,3:4) = [Results.std(l+1:l+EstimOpt.CutOff),pv(Results.bhat(l+1:l+EstimOpt.CutOff),Results.std(l+1:l+EstimOpt.CutOff))];


Results.R = [Results.DetailsA1;Results.DetailsV1; Results.DetailsA2;Results.DetailsV2; Results.DetailsCutOffs];
EstimOpt.params = length(Results.bhat);

if isfield(EstimOpt,'BActive')
    EstimOpt.params = EstimOpt.params - sum(EstimOpt.BActive == 0);
end
%  Results.stats = [Results.LL;Results.LL0;1 - Results.LL/Results.LL0;R2;((2*EstimOpt.params - 2*Results.LL))/EstimOpt.NObs;((log(EstimOpt.NObs)*EstimOpt.params - 2*Results.LL))/EstimOpt.NObs;EstimOpt.NObs;EstimOpt.NP;EstimOpt.params];
Results.stats = [Results.LL; nan(8,1)];
Results.EstimOpt = EstimOpt;
Results.OptimOpt = OptimOpt;
Results.INPUT = INPUT;

Head = cell(1,2);

if EstimOpt.FullCov == 0
    Head(1,1) = {'HMXL_FSE_d'};
else
    Head(1,1) = {'HMXL_FSE'};
end

% Template1 = {'DetailsA1','DetailsV1','DetailsInter', 'DetailsA2','DetailsV2'};
Template2 = cell(3,2);
Template2(1,:) = {'DetailsA1','DetailsV1'};
Template2{2,1} = 'DetailsInter';
Template2(3,:) =  {'DetailsA2','DetailsV2'};
% Template2{4,1} =  'DetailsCutOffs';
Template1 = Template2;
Names.DetailsA1 = EstimOpt.NamesA1;
Heads.DetailsA1 = {'Choice model: Means';'tc'};
Heads.DetailsV1 = {'Standard Deviations';'lb'};
ST = [];

Heads.DetailsInter(:,2) = [EstimOpt.NamesA2(EstimOpt.Dist2 ~= -1);{'lb'}];
Heads.DetailsInter(1:2,1) = {'Interactions of means with LV';'lc'};
Names.DetailsInter = EstimOpt.NamesA1;
 ST = [ST, {'DetailsInter'}];

 if EstimOpt.NVarStr > 0
    for i = 1:EstimOpt.NVarRnd2
        Results.SE(1:EstimOpt.NVarStr,4*i-3:4*i) = Results.DetailsStr((i-1)*EstimOpt.NVarStr+1:i*EstimOpt.NVarStr,:);
        Heads.SE(1:2,1) = {'Structural equations';'lc'};
        Heads.SE(i,2) = EstimOpt.NamesLV(i);
    end
    Heads.SE(EstimOpt.NVarRnd2+1,2) = {'lb'};
    %Results.R_out(EstimOpt.NVarA+7+(2 + EstimOpt.NVarS)*(EstimOpt.NVarS>0):EstimOpt.NVarA+6+EstimOpt.NVarStr+(2 + EstimOpt.NVarS)*(EstimOpt.NVarS>0),2 + 3*(i-1):1+3*i) = num2cell(Results.DetailsS((i-1)*EstimOpt.NVarStr+1:i*EstimOpt.NVarStr,:));
end
if isfield(Results,'SE')
    Temp1 = cell(1, size(Template1,2));
    Temp1(1,1) = {'SE'};
    Template1 = [Template1;Temp1];
    Temp2 = cell(1, size(Template2,2));
    Temp2(1,1) = {'SE'};
    Template2 = [Template2;Temp2];
    Names.SE = EstimOpt.NamesStr;
    ST = [ST, {'SE'}];
end
Names.DetailsA2 = EstimOpt.NamesA2;
Heads.DetailsA2 = {'FSE model: Means';'tc'};
Heads.DetailsV2 = {'Standard Deviations';'lb'};
Names.DetailsA2 = EstimOpt.NamesA2;
 ST = [ST, {'DetailsA2'}];

% if EstimOpt.NVarM > 0
%     Template1 = [Template1,'DetailsCM'];
%     Temp2 = cell(1,size(Template2,2));
%     Temp2(1,1) = {'DetailsCM'};
%     Template2 = [Template2; Temp2];
%     Heads.DetailsCM(:,2) = [EstimOpt.NamesM;{'lb'}];
%     Heads.DetailsCM(1:2,1) = {'Interactions of means';'lc'};
% end
% 
% if EstimOpt.NVarS >0
%     Temp1 = cell(1,size(Template1,2));
%     Temp1(1,1) = {'DetailsScale'};
%     Template1 = [Template1; Temp1];
%     Temp2 = cell(1,size(Template2,2));
%     Temp2(1,1) = {'DetailsScale'};
%     Template2 = [Template2; Temp2];
%     Names.DetailsScale = EstimOpt.NamesS;
%     Heads.DetailsS = {'Covariates of Scale';'tb'};
%     ST = {'DetailsScale'};
% end
% 
% if EstimOpt.NVarStr > 0
%     for i = 1: EstimOpt.NLatent
%         Results.SE(1:EstimOpt.NVarStr,4*i-3:4*i) = Results.DetailsS((i-1)*EstimOpt.NVarStr+1:i*EstimOpt.NVarStr,:);
%         Heads.SE(1:2,1) = {'Structural equations';'lc'};
%         Heads.SE(i,2) = EstimOpt.NamesLV(i);
%     end
%     Heads.SE(EstimOpt.NLatent+1,2) = {'lb'};
%     %Results.R_out(EstimOpt.NVarA+7+(2 + EstimOpt.NVarS)*(EstimOpt.NVarS>0):EstimOpt.NVarA+6+EstimOpt.NVarStr+(2 + EstimOpt.NVarS)*(EstimOpt.NVarS>0),2 + 3*(i-1):1+3*i) = num2cell(Results.DetailsS((i-1)*EstimOpt.NVarStr+1:i*EstimOpt.NVarStr,:));
% end
% 
% if isfield(Results,'SE')
%     Temp1 = cell(1, size(Template1,2));
%     Temp1(1,1) = {'SE'};
%     Template1 = [Template1;Temp1];
%     Temp2 = cell(1, size(Template2,2));
%     Temp2(1,1) = {'SE'};
%     Template2 = [Template2;Temp2];
%     Names.SE = EstimOpt.NamesStr;
%     ST = [ST, {'SE'}];
% end
% %l = EstimOpt.NVarA+3+(2 + EstimOpt.NVarS)*(EstimOpt.NVarS>0); % this is for indexing in R_out
% k = 0;
% if any(EstimOpt.MeaSpecMatrix == 2)
%     Results.DetailsOP = [];
% end
% 
% 
%% Tworzenie stopki


Tail = cell(17,2);
Tail(2,1) = {'Model diagnostics'};
Tail(3:17,1) = {'LL at convergence';'LL at constant(s) only';strcat('McFadden''s pseudo-R',char(178));strcat('Ben-Akiva-Lerman''s pseudo-R',char(178));'AIC/n';'BIC/n';'n (observations)';'r (respondents)';'k (parameters)';' ';'Estimation method';'Simulation with';'Optimization method';'Gradient';'Hessian'};
Tail(3,2) = num2cell(Results.LL);
Tail(4:11,2) = num2cell(nan);
% if isfield(Results_old,'MNL0') && isfield(Results_old.MNL0,'LL')
%     Tail(3:11,2) = num2cell(Results.stats);
% end

if any(INPUT.W ~= 1)
    Tail(13,2) = {'weighted simulated maximum likelihood'};
else
    Tail(13,2) = {'simulated maximum likelihood'};
end

switch EstimOpt.Draws
    case 1
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','pseudo-random draws']};
    case 2
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','Latin Hypercube Sampling draws']};
    case  3
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','Halton draws (skip = ',num2str(EstimOpt.HaltonSkip), '; leap = ',num2str(EstimOpt.HaltonLeap),')']};
    case 4
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','Halton draws with reverse radix scrambling (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),')']};
    case 5
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','Sobol draws (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),')']};
    case 6
        Tail(14,2) = {[num2str(EstimOpt.NRep),' ','Sobol draws with random linear scramble and random digital shift (skip = ',num2str(EstimOpt.HaltonSkip),'; leap = ',num2str(EstimOpt.HaltonLeap),')']};
end

Tail(15,2) = {OptimOpt.Algorithm;};

if strcmp(OptimOpt.GradObj,'on')
    if EstimOpt.NumGrad == 0
        Tail(16,2) = {'user-supplied, analytical'};
    else
        Tail(16,2) = {['user-supplied, numerical ',num2str(OptimOpt.FinDiffType)]};
    end
else
    Tail(16,2) = {['built-in, ',num2str(OptimOpt.FinDiffType)]};
    
end

% outHessian = [];
if isequal(OptimOpt.Algorithm,'quasi-newton')
    outHessian='off, ';
    switch EstimOpt.HessEstFix
        case 0
            outHessian = [outHessian,'retained from optimization'];
        case 1
            outHessian = [outHessian,'ex-post calculated using BHHH'];
        case 2
            outHessian = [outHessian,'ex-post calculated using high-precision BHHH'];
        case 3
            outHessian = [outHessian,'ex-post calculated numerically'];
        case 4
            outHessian = [outHessian,'ex-post calculated analytically'];
    end
else
    if strcmp(OptimOpt.Hessian,'user-supplied')
        if EstimOpt.ApproxHess == 1
            outHessian = 'user-supplied, BHHH, ';
        else
            outHessian = 'user-supplied, analytical, ';
        end
    else
        outHessian = ['built-in, ',num2str(OptimOpt.HessUpdate),', '];
    end
    switch EstimOpt.HessEstFix
        case 0
            outHessian = [outHessian,'retained from optimization'];
        case 1
            outHessian = [outHessian,'ex-post calculated using BHHH'];
        case 2
            outHessian = [outHessian,'ex-post calculated using high-precision BHHH'];
        case 3
            outHessian = [outHessian,'ex-post calculated numerically'];
        case 4
            outHessian = [outHessian,'ex-post calculated analytically'];
    end
end

Tail(17,2) = {outHessian};

%% Tworzenie ResultsOut, drukowanie na ekran i do pliku .xls
% 
if EstimOpt.Display ~= 0
    Results.Dist = EstimOpt.Dist1;
    Results.R_out = genOutput(EstimOpt, Results, Head, Tail, Names, Template1, Template2, Heads, ST);    
end